/*
  Macro           DirSize
  Author          Carlo Hogeveen
  Website         eCarlo.nl/tse
  Version         v1.1.2   4 Nov 2023
  Compatibility   TSE 4.41.44 (18 Jan 2021) upwards, all variants


  This tool creates a list of directories sorted on decreasing size.

  Its input is a string of one or more drives or directories.

  Its output is a list of the input directories and their subdirectories,
  prefixed by their size, number of files, date, and time,
  and sorted on decreasing size and number of files.


  INSTALLATION

  Copy this file to TSE's "mac" directory and compile it there, for example by
  opening it there in TSE and applying the menu Macro, Compile.


  USAGE

  To run it just execute the tool as a macro, for example by applying the
  menu Macro -> Execute, and entering "DirSize".

  You can add parameters after the macro name or when the tool prompts for
  input.

  My advice is to add no more than one drive or directory per run.
  In my case this results in the most significant output.

  But you can enter as many drives and directories as you like,
  or even the single parameter "all" to list all drives.

  Directory names containing a space need to be quoted.

  A trailing slash is allowed, not required.
  Unlike Windows' default functionality, even "C:" is assumed to mean "C:\".

  If "all" is the only parameter then it stands for all root directories
  with a Windows drive letter, like "a: b: c: d: e: ... x: y: z:".
  I do not advise "all" in combination with a slow and/or huge network drive.


  DETAILS

  For DirSize a directory's size is the sum of the logical sizes of its files,
  including those in subdirectories.

  A file's logical size is how many bytes it contains.
  It does not include the disk space overhead of the operating system.
  For example, a directory itself takes up a little space for references to
  subdirectories, files, and their properties. That space is not counted.
  Also, an operating system typically allocates each file's disk space in
  blocks of a fixed size. For Example, if such a disk block is 4096 bytes and a
  file contains 5000 bytes, then the file's logical size is 5000 bytes, while
  the space it takes on disk is 2 * 4096 = 8192 bytes.


  KNOWN ERROR

  DirSize intentionally slows down for network drives, but it cannot determine
  which drives are network drives.
  The reason DirSize slows down for network drives is two-fold:
  - I have noticed that TSE can send network requests faster than a network can
    handle, causing errors.
  - Given that a network can be used concurrently by multiple users/computers,
    I decided that TSE should not hog a network drive.
  The bad part is, that because DirSize cannot determine which drives are
  network drives (yet?), it currently blindly assumes that directories starting
  with "c", "d" or "/" are local and not to be slowed down for, and that all
  other directories are on a network and to be slowed down for.


  EXCEPTIONS

  Windows automatically converts directory names with application-incompatible
  characters to the application's closest compatible equivalent.
  This might cause a directory to not be listed by its real name.
  For example, a directory listed as "The a test results" might actually have a
  name with the Greek alpha character instead of an "a".

  There are four known cases in which a directory or file is not counted or
  wrongly counted. They are explained in more detail further on.
  Here is a short overview:
  - If the full name of a parent directory has a too deep path, then its child
    directories and files will not be counted.
  - If the full name of a directory contains TSE-incompatible characters, then
    its subdirectories and files will not be counted.
  - On 64-bit systems Windows hides some system directories and files from
    32-bit programs like TSE, for example:
      "C:\Windows\System32\drivers\etc\hosts".
  - Windows NTFS' alternate data streams (ADS) are not counted.

  This tool mitigates some of these flaws by marking such directories or their
  parent directories as "inaccessible".


  TOO DEEP PATHS

  Because TSE's string limit is 255 characters and because TSE uses the old
  Windows ANSI APIs which are likewise string limited, TSE can list the
  contents of a directory if its full path has no more than 255 characters.
  The first consequence is, that such a directory itself is inaccessible
  to TSE, and that all its contents are invisible to TSE.
  The second consequence is, that TSE can still "see" the name of a child
  directory or file with a full name of more than 255 characters if the full
  name of their parent directory is less than 255 characters, but that TSE can
  not access such child directories and files.


  TSE-INCOMPATIBLE CHARACTERS IN NAME

  Windows directory and file names are stored in the UTF-16LE character
  encoding, which is one of the Unicode character encodings.
  Unicode 15.1 (12 Sep 2023) supports 149,813 characters.
  TSE at best supports ANSI (a.k.a. "Windows-1252") which has 218 characters.
  TSE uses the old Windows ANSI APIs to access Windows directories and files.
  This has the benefit that if a Unicode character has an ANSI equivalent,
  then the Windows file APIs convert the character back and forth.
  For instance, the Unicode character "latin small letter e with acute" exists
  in ANSI too, and the Windows API makes TSE think it reads and write the ANSI
  character for directory and file names.
  For Unicode characters without an ANSI equivalent the Windows API converts
  the Unicode character one-way to a close ANSI alternative for TSE to see.
  For instance, in directory and file names TSE sees the Greek letter alpha and
  other close "foreign" (to ANSI) characters as an "a".
  But having seen that, TSE can not open such a directory or file because its
  actual name does not contain an "a", and the low-level Windows API does not
  know, remember, or want to guess which foreign character that "a" came from.
  So, for Windows directory names that contain a Unicode character that has no
  ANSI equivalent DirSize will show a close alternative character and mark the
  directory in the list as "inaccessible".

  TSE's Console version does not even support ANSI, so while some equivalent
  things might happen, it is beyond the scope of this tool to describe the many
  possibilities.


  INVISIBLE 64-BIT DIRECTORIES AND FILES

  This problem only exists on 64-bit systems.
  For backwards compatibility both 32-bit and 64-bit programs can and
  should access "C:\Windows\System32" to get to Windows programs and files
  that perform Windows tasks.
  But typically those programs and files differ for 32-bit and 64-bit.
  Windows solves this as follows:
  - It maintains two directories:
    "C:\Windows\System32" for 64-bit programs and
    "C:\Windows\SysWOW64" for 32-bit programs.
  - Whenever a 32-bit program (like TSE) thinks it accesses the directory
    "C:\Windows\System32" Windows invisibly substitutes it with the
    directory "C:\Windows\SysWOW64".
  Presumably because Windows itself is 64-bit, directories and files that
  should only be relevant to Windows are only maintained in
  "C:\Windows\System32", but there they are invisible for 32-bit programs
  like TSE.
  DirSize runs from TSE, so it is 32-bit too. It does not even see any 64-bit
  directories and files, so it can not even report them as "inaccessible".

  Example:
  TSE can not open "C:\" and browse to the file
  "C:\Windows\System32\drivers\etc\hosts", because TSE is a 32-bit program and
  the "etc" directory does not exist in the "C:\Windows\SysWOW64\drivers"
  directory.
  It only exists in the real "C:\Windows\System32\drivers" directory, which TSE
  can not see.
  If the full filename is supplied, then Windows gets smart and falls back to
  the real directory and lets TSE open but not write the file.
  If TSE is run as administrator AND the full filename is supplied,
  then TSE can open and write the file.


  TODO
    MUST
    SHOULD
    - I plan to research whether DirList can use Windows APIs to also access
      files in deep paths, files with non-TSE characters in their names, and
      alternate data streams.
      The result of that research should also apply to DirSize.
    COULD
    WOULD


  HISTORY
    v0.0.0.1   2 Nov 2023
      Initial alpha release.

    v1         3 Nov 2023
      Initial user version.

    v1.1       3 Nov 2023
      Per request added a column with the number of files.
      Added a better abort message for a non-existing input directory.

    v1.1.1     4 Nov 2023
      Fixed that DirSize was slow for Linux.
      Documented a known error, caused by DirSize's inability to determine
      which drives are network drives.
      Fixed that the created list's name could contain a space.

    v1.1.2     4 Nov 2023
      Because a user reported DirSize crashing TSE for a certain input
      directory on their Linux system, I added the capability to DirSize to
      write its progress to a debug log file.
*/



/*
  T E C H N I C A L   N O T E S


  ALGORITHM FOR CREATING THE LIST

  Given:
    A string of input directories.
    An empty current buffer in which to create the list.

  Main algorithm (with details left out):
    For each directory in the input string do "process the directory".
    Sort the end result list.

  Process the directory (with details left out):
    Add an empty line after the currenty line, go to it,
    and fill it with the directory's date, time, and name.
    Initialize a variable dir_size with "0".
    For each subdirectory or file in the directory:
      If it is a sub directory, then do "process the directory" for the
      subdirectory and add its dir_size to our dir_size.
      If it is a file, then add the file's size to our dir_size.
    Insert our dir_size at the beginning of the line.
    Go one line up.
    Return our dir_size.

  Note:
    In theory Windows does not garantee returning a directory's subdirectories
    and files in any order. In practice it does sort subdirectories and files
    on name. Despite that, the above algorithm creates a horribly sorted list.
    For the user this does not matter, because we sort the end result anyway.

*/





/*
  Debugging
  The global variable DEBUG must be defined whether you want to debug or not.
  Its value must be TRUE or FALSE.
  If DEBUG is set to TRUE, then debug information will be written to the file
  "TSE_Debug_<macro name>_<time in hhmmsshh>" in "the temporary directory".
  In Windows "the temporary directory" is determined by the environment
  variable TMP, in Linux it is the "/tmp" directory.
*/
#define DEBUG FALSE





// Compatibility restrictions and mitigations

#ifdef LINUX
  #define WIN32 FALSE
#endif

#ifndef WIN32
   16-bit versions of TSE are not supported. You need at least TSE 4.41.44.
#endif

#ifndef EDITOR_VERSION
   Editor Version is older than TSE 3.0. You need at least TSE 4.41.44.
#endif

#ifndef INTERNAL_VERSION
   Editor Version is older than TSE 4.41.44. You need at least TSE 4.41.44.
#endif


// End of compatibility restrictions and mitigations





// Start of debug log implementation


#if DEBUG

  integer debug_handle = -1


  integer proc debug_write(string msg)
    integer old_DateFormat = Set(DateFormat, 6)
    integer old_timeFormat = Set(TimeFormat, 1)
    integer ok             = TRUE

    #ifdef LINUX
      string NEWLINE [1] = Chr(10)
    #else
      string NEWLINE [2] = Chr(13) + Chr(10)
    #endif

    if debug_handle == -1
      Warn('No open debug log to write to.')
      ok = FALSE
    else
      if fWrite(debug_handle, GetDateStr() + ' ' + GetTimeStr() + ' ') == -1
        Warn('Debug log write date time prefix failed.')
        ok = FALSE
      else
        if fWrite(debug_handle, msg) == -1
          Warn('Debug log write message failed.')
          ok = FALSE
        else
          if fWrite(debug_handle, NEWLINE) == -1
            Warn('Debug log write newline failed.')
            ok = FALSE
          endif
        endif
      endif
    endif
    Set(DateFormat, old_DateFormat)
    Set(TimeFormat, old_timeFormat)
    return(ok)
  end debug_write


  integer proc debug_open()
    string  log_fqn [MAXSTRINGLEN] = 'TSE_Debug_' +
                                     SplitPath(CurrMacroFilename(), _NAME_) +
                                     '_' + Format(GetTime():7:'0') + '.log'
    integer log_attrs              = 0
    integer ok                     = TRUE

    #ifdef LINUX
      log_fqn = '/tmp/' + log_fqn
    #else
      log_fqn = GetEnvStr('TMP') + '\' + log_fqn
    #endif

    if debug_handle <> -1
      Warn('Debug log already open.')
    else
      log_attrs = FileExists(log_fqn)
      if log_attrs
        if log_attrs & _DIRECTORY_
          Warn('Debug cannot log to directory.')
          ok = FALSE
        else
          debug_handle = fOpen(log_fqn, _OPEN_READWRITE_)
          if debug_handle == -1
            Warn('Debug open log failed.')
            ok = FALSE
          else
            if fSeek(debug_handle, 0, _SEEK_END_) == -1
              Warn('Debug first log fSeek failed.')
              ok = FALSE
            else
              debug_write('Debug log reopened.')
            endif
          endif
        endif
      else
        if FileExists(SplitPath(log_fqn, _DRIVE_|_PATH_)) & _DIRECTORY_
          debug_handle = fCreate(log_fqn)
          if debug_handle == -1
            Warn('Debug creating log failed.')
            ok = FALSE
          else
            ok = debug_write('Debug log opened.')
          endif
        else
          Warn('Debug cannot log to non-existing directory.')
          ok = FALSE
        endif
      endif
    endif

    return(ok)
  end debug_open


  integer proc debug_close()
    integer ok = FALSE
    if debug_handle <> -1
      ok = debug_write('Debug log closed.')
      if fClose(debug_handle)
        debug_handle = -1
      else
        Warn('Debug failed to close log.')
        ok = FALSE
      endif
    endif
    return(ok)
  end debug_close

#endif


// End of debug log implementation.





// START OF BIGINT IMPLEMENTATION

forward string  proc add_bigints (string term1, string term2)
forward string  proc subtract_bigint_minus_bigint (string minuend,
                                                   string subtrahend)
forward string  proc multiply_bigints (string factor1, string factor2)
forward string  proc divide_bigint_by_bigint (string dividend, string divisor)
forward integer proc compare_bigints
                                  (string left_quantity, string right_quantity)


// START OF INTERNAL FUNCTIONS - DO NOT CALL THESE

proc standardize_bigint(    string  input_bigint,
                        var string  output_positive_bigint,
                        var integer output_bigint_is_negative)
  integer i = 0
  // Remove surrounding white space and initialize the output sign as positive.
  output_positive_bigint    = Trim(input_bigint)
  output_bigint_is_negative = FALSE
  // Remove leading signs and remaining leading spaces,
  // toggling the output sign for each minus sign.
  while (output_positive_bigint[1] in '+', '-', ' ')
    if output_positive_bigint[1] == '-'
      output_bigint_is_negative = not output_bigint_is_negative
    endif
    output_positive_bigint = output_positive_bigint[2:MAXSTRINGLEN]
  endwhile
  // Remove trailing signs and remaining trailing spaces,
  // toggling the output sign for each minus sign.
  while (   output_positive_bigint[Length(output_positive_bigint)]
         in '+', '-', ' ')
    if output_positive_bigint[Length(output_positive_bigint)] == '-'
      output_bigint_is_negative = not output_bigint_is_negative
    endif
    output_positive_bigint =
      output_positive_bigint[1: Length(output_positive_bigint) - 1]
  endwhile
  // Delete group separators, like thousands separators, but deleting them
  // anywhere between two digits to account for all countries and styles.
  // Any decimal separators (commas or points) are not recognized as such and
  // are deleted when they resemble group separators.
  i = 2
  while i < Length(output_positive_bigint)
    if  (output_positive_bigint[i    ] in ',', '.', ' ', "'", '_')
    and (output_positive_bigint[i - 1] in '0' .. '9'             )
    and (output_positive_bigint[i + 1] in '0' .. '9'             )
      output_positive_bigint = output_positive_bigint[1     : i - 1       ]
                             + output_positive_bigint[i + 1 : MAXSTRINGLEN]
    else
      i = i + 1
    endif
  endwhile
  // Set output to an empty string if an illegal character is found.
  for i = 1 to Length(output_positive_bigint)
    if not (output_positive_bigint[i] in '0' .. '9')
      output_positive_bigint = ''
      break
    endif
  endfor
  // Remove leading zeros except the last one if the output value is zero.
  while output_positive_bigint[1] == '0'
  and   Length(output_positive_bigint) > 1
    output_positive_bigint = output_positive_bigint[2:MAXSTRINGLEN]
  endwhile
  // If the output value is zero, then make it not negative.
  if output_positive_bigint == '0'
    output_bigint_is_negative = FALSE
  endif
end standardize_bigint

integer proc compare_positive_bigints(string left_quantity, string right_quantity)
  integer result = 0
  if Length(left_quantity) > Length(right_quantity)
    result = 1
  elseif Length(left_quantity) < Length(right_quantity)
    result = -1
  elseif left_quantity > right_quantity
    result = 1
  elseif left_quantity < right_quantity
    result = -1
  endif
  return(result)
end compare_positive_bigints

string proc add_positive_bigints(string p_term1, string p_term2)
  integer carry                       = 0
  integer i                           = 0
  integer last_digit                  = 0
  integer len                         = 0
  integer sum_of_two_digits_and_carry = 0
  string  term1        [MAXSTRINGLEN] = p_term1
  string  term2        [MAXSTRINGLEN] = p_term2
  string  total_sum    [MAXSTRINGLEN] = ''
  if  p_term1 <> ''
  and p_term2 <> ''
    if Length(term1) > Length(term2)
      len   = Length(term1)
      term2 = Format(term2:len:'0')
    else
      len   = Length(term2)
      term1 = Format(term1:len:'0')
    endif
    for i = len downto 1
      sum_of_two_digits_and_carry = Val(term1[i]) + Val(term2[i]) + carry
      last_digit                  = sum_of_two_digits_and_carry mod 10
      carry                       = sum_of_two_digits_and_carry  /  10
      total_sum                   = Str(last_digit) + total_sum
    endfor
    if carry
      if Length(total_sum) == MAXSTRINGLEN
        total_sum = ''
      else
        total_sum = Str(carry) + total_sum
      endif
    endif
  endif
  return(total_sum)
end add_positive_bigints

string proc subtract_positive_larger_minus_smaller_bigint(string p_minuend,
                                                          string p_subtrahend)
  string  minuend      [MAXSTRINGLEN] = p_minuend
  string  subtrahend      [MAXSTRINGLEN] = p_subtrahend
  integer i                     = 0
  integer j                     = 0
  integer len                   = 0
  integer keep_borrowing        = FALSE
  string  difference [MAXSTRINGLEN] = ''
  if  p_minuend    <> ''
  and p_subtrahend <> ''
    if Length(minuend) > Length(subtrahend)
      len        = Length(minuend)
      subtrahend = Format(subtrahend:len:'0')
    else
      len     = Length(subtrahend)
      minuend = Format(minuend:len:'0')
    endif
    for i = len downto 1 by 1
      if     minuend[i] > subtrahend[i]
        difference     = Str(Val(minuend[i]) - Val(subtrahend[i]))
                       + difference
      elseif minuend[i] < subtrahend[i]
        difference     = Str(Val(minuend[i]) + 10 - Val(subtrahend[i]))
                       + difference
        j              = i - 1
        keep_borrowing = TRUE
        while j > 0
        and   keep_borrowing
          if minuend[j] == '0'
            minuend[j]     = '9'
          else
            minuend[j]     = Str(Val(minuend[j]) - 1)
            keep_borrowing = FALSE
          endif
          j = j - 1
        endwhile
      else
        difference = '0' + difference
      endif
    endfor
    while difference[1]     == '0'
    and   Length(difference) >  1
      difference = difference[2:MAXSTRINGLEN]
    endwhile
  endif
  return(difference)
end subtract_positive_larger_minus_smaller_bigint

/*
  For multiplying I use an algorith of my own that keeps intermediate results
  small. This turns out to be ideal for TSE's small integer size.

  It is based on this principle:
  Call each right-most digit a ones-digit, each digit to the left of that a
  tens-digit, each digit to the left of that a hundreds-digit, etcetera.
  For this example assume two three-digit factors to multiply to a product.
  The ones-digit of the product is only determined by the multiplication of the
  ones-digits of the two factors, and also results in a carry.
  The tens-digit of the product is determined by adding the carry plus the
  multiplications of the ones-digits of one factor with the tens-digits
  of the other factor, again resulting in a new carry.
  The hundreds-digit of the product is determined by adding the carry plus the
  multiplications of the ones-digits of one factor with the hundreds-digits
  of the other factor plus the multiplication of the tens-digits of both
  factors, again resulting in a new carry.
  The thousands-digit of the product is determined by adding the carry plus the
  multiplications of the tens-digit of one factor with the hundreds-digit of
  the other factor, again resulting in a new carry.
  The ten-thousands-digit of the product is determined by adding the carry plus
  the multiplication of the hundreds-digits of the factors, again resulting in
  a new carry.
  The hundred-thousands-digit of the product is determined by the final carry.

  For example
      123
      456 *
      ---
    56088
  can be calculated as follows
        6 * 3                 = 8 carry 1
    1 + 6 * 2 + 5 * 3         = 8 carry 2
    2 + 6 * 1 + 5 * 2 + 4 * 3 = 0 carry 3
    3 + 5 * 1 + 4 * 2         = 6 carry 1
    1 + 4 * 1                 = 5 carry 0
*/

string proc multiply_positive_bigints(string p_factor1, string p_factor2)
  integer carry                  = 0
  string  factor1 [MAXSTRINGLEN] = p_factor1
  string  factor2 [MAXSTRINGLEN] = p_factor2
  integer factors_left_pos       = 0
  integer factors_right_pos      = 0
  integer factor1_pos            = 0
  integer factor2_pos            = 0
  integer len                    = 0
  string  product [MAXSTRINGLEN] = ''
  integer product_pos_digit      = 0
  integer product_pos_product    = 0
  integer size_error             = FALSE
  if Length(factor1) > Length(factor2)
    len     = Length(factor1)
    factor2 = Format(factor2:len:'0')
  else
    len     = Length(factor2)
    factor1 = Format(factor1:len:'0')
  endif
  factors_left_pos = len
  for factors_right_pos = len downto 1 by 1
    for factors_left_pos = factors_left_pos downto 1
      if Length(product) == MAXSTRINGLEN
        size_error = TRUE
      else
        product_pos_product = 0
        for factor1_pos = factors_right_pos downto factors_left_pos
          factor2_pos         = factors_left_pos
                              + (factors_right_pos - factor1_pos)
          product_pos_product = product_pos_product
                              + Val(factor1[factor1_pos])
                              * Val(factor2[factor2_pos])
        endfor
        product_pos_product = product_pos_product + carry
        carry               = product_pos_product / 10
        product_pos_digit   = product_pos_product mod 10
        product             = Str(product_pos_digit) + product
      endif
    endfor
    factors_left_pos = 1
  endfor
  if carry
    if Length(product) == MAXSTRINGLEN
      size_error = TRUE
    else
      product = Str(carry) + product
    endif
  endif
  if size_error
    product = ''
  endif
  while product[1]     == '0'
  and   Length(product) >  1
    product = product[2:MAXSTRINGLEN]
  endwhile
  return(product)
end multiply_positive_bigints

/*
  The algorithm used is ordinary long division.
  Division is an expensive operation.
  A small performance improvement is realized by storing the results of some
  multiplications that are likely to recur for the same division.
*/

string bigint_pre_multiplied_divisor [MAXSTRINGLEN] = ''
string bigint_divisor_times_0        [MAXSTRINGLEN] = ''
string bigint_divisor_times_1        [MAXSTRINGLEN] = ''
string bigint_divisor_times_2        [MAXSTRINGLEN] = ''
string bigint_divisor_times_3        [MAXSTRINGLEN] = ''
string bigint_divisor_times_4        [MAXSTRINGLEN] = ''
string bigint_divisor_times_5        [MAXSTRINGLEN] = ''
string bigint_divisor_times_6        [MAXSTRINGLEN] = ''
string bigint_divisor_times_7        [MAXSTRINGLEN] = ''
string bigint_divisor_times_8        [MAXSTRINGLEN] = ''
string bigint_divisor_times_9        [MAXSTRINGLEN] = ''
string bigint_divisor_times_10       [MAXSTRINGLEN] = ''

string proc get_bigint_divisor_product(string  p_divisor,
                                       integer p_number_of_times)
  string product [MAXSTRINGLEN] = ''
  if p_divisor <> bigint_pre_multiplied_divisor
    bigint_pre_multiplied_divisor = p_divisor
    bigint_divisor_times_0        = '0'
    bigint_divisor_times_1        = p_divisor
    bigint_divisor_times_2        = ''
    bigint_divisor_times_3        = ''
    bigint_divisor_times_4        = ''
    bigint_divisor_times_5        = ''
    bigint_divisor_times_6        = ''
    bigint_divisor_times_7        = ''
    bigint_divisor_times_8        = ''
    bigint_divisor_times_9        = ''
    if Length(p_divisor) <> MAXSTRINGLEN
      bigint_divisor_times_10     = p_divisor + '0'
    endif
  endif
  case p_number_of_times
    when  0 product = bigint_divisor_times_0
    when  1 product = bigint_divisor_times_1
    when  2 product = bigint_divisor_times_2
    when  3 product = bigint_divisor_times_3
    when  4 product = bigint_divisor_times_4
    when  5 product = bigint_divisor_times_5
    when  6 product = bigint_divisor_times_6
    when  7 product = bigint_divisor_times_7
    when  8 product = bigint_divisor_times_8
    when  9 product = bigint_divisor_times_9
    when 10 product = bigint_divisor_times_10
  endcase
  if product == ''
    product = multiply_positive_bigints(p_divisor, Str(p_number_of_times))
    case p_number_of_times
      when  0 bigint_divisor_times_0  = product
      when  1 bigint_divisor_times_1  = product
      when  2 bigint_divisor_times_2  = product
      when  3 bigint_divisor_times_3  = product
      when  4 bigint_divisor_times_4  = product
      when  5 bigint_divisor_times_5  = product
      when  6 bigint_divisor_times_6  = product
      when  7 bigint_divisor_times_7  = product
      when  8 bigint_divisor_times_8  = product
      when  9 bigint_divisor_times_9  = product
      when 10 bigint_divisor_times_10 = product
    endcase
  endif
  return(product)
end get_bigint_divisor_product

string proc divide_positive_bigint_by_bigint(string p_dividend,
                                             string p_divisor)
  integer dividend_digit_index                     = 0
  string  dividend_part        [MAXSTRINGLEN] = ''
  string  divisor_product      [MAXSTRINGLEN] = ''
  integer number_of_times                     = 0
  string  prev_divisor_product [MAXSTRINGLEN] = ''
  string  quotient             [MAXSTRINGLEN] = ''
  for dividend_digit_index = 1 to Length(p_dividend)
    dividend_part   = iif(dividend_part == '0', '', dividend_part)
                    + p_dividend[dividend_digit_index]
    number_of_times =  0
    divisor_product = '0'
    repeat
      number_of_times      = number_of_times + 1
      prev_divisor_product = divisor_product
      divisor_product      = get_bigint_divisor_product(p_divisor,
                                                        number_of_times)
    until number_of_times                                          == 10
       or compare_positive_bigints(divisor_product, dividend_part) ==  1
    dividend_part = subtract_positive_larger_minus_smaller_bigint(
                                          dividend_part, prev_divisor_product)
    quotient      = quotient + Str(number_of_times - 1)
  endfor
  while quotient[1]     == '0'
  and   Length(quotient) >  1
    quotient = quotient[2:MAXSTRINGLEN]
  endwhile
  return(quotient)
end divide_positive_bigint_by_bigint

proc bigint_compiler_pacifier()
  integer i    = 0
  string  s[1] = ''
  if FALSE
    bigint_compiler_pacifier()
    standardize_bigint(s, s, i)
    compare_positive_bigints(s, s)
    add_positive_bigints(s, s)
    subtract_positive_larger_minus_smaller_bigint(s, s)
    multiply_positive_bigints(s, s)
    get_bigint_divisor_product(s, i)
    divide_positive_bigint_by_bigint(s, s)

    add_bigints(s, s)
    subtract_bigint_minus_bigint(s, s)
    multiply_bigints(s, s)
    divide_bigint_by_bigint(s, s)
    compare_bigints(s, s)
  endif
end bigint_compiler_pacifier

// END OF INTERNAL FUNCTIONS - DO NOT CALL THESE





// START OF SAFE FUNCTIONS - THESE CAN BE CALLED BY YOUR OWN CODE

string proc add_bigints(string p_term1, string p_term2)
  string  sum   [MAXSTRINGLEN] = ''
  string  term1 [MAXSTRINGLEN] = ''
  integer term1_is_negative    = FALSE
  string  term2 [MAXSTRINGLEN] = ''
  integer term2_is_negative    = FALSE
  standardize_bigint(p_term1, term1, term1_is_negative)
  standardize_bigint(p_term2, term2, term2_is_negative)
  if  term1 <> ''
  and term2 <> ''
    if term1_is_negative == term2_is_negative
      sum = add_positive_bigints(term1, term2)
      if  sum <> ''
      and term1_is_negative
        sum = iif(Length(sum) == MAXSTRINGLEN, '', '-' + sum)
      endif
    else
      case compare_positive_bigints(term1, term2)
        when 0
          sum = '0'
        when -1
          sum = subtract_positive_larger_minus_smaller_bigint(term2, term1)
          if  sum <> ''
          and term2_is_negative
            sum = iif(Length(sum) == MAXSTRINGLEN, '', '-' + sum)
          endif
        when 1
          sum = subtract_positive_larger_minus_smaller_bigint(term1, term2)
          if  sum <> ''
          and term1_is_negative
            sum = iif(Length(sum) == MAXSTRINGLEN, '', '-' + sum)
          endif
      endcase
    endif
  endif
  return(sum)
end add_bigints

string proc subtract_bigint_minus_bigint(string p_minuend, string p_subtrahend)
  string  difference [MAXSTRINGLEN] = ''
  string  minuend    [MAXSTRINGLEN] = ''
  integer minuend_is_negative       = FALSE
  string  subtrahend [MAXSTRINGLEN] = ''
  integer subtrahend_is_negative    = FALSE
  standardize_bigint(p_minuend   , minuend   ,    minuend_is_negative)
  standardize_bigint(p_subtrahend, subtrahend, subtrahend_is_negative)
  if  minuend    <> ''
  and subtrahend <> ''
    if Length(Trim(subtrahend)) == MAXSTRINGLEN
      difference = ''
    else
      difference = add_bigints(minuend, '-' + subtrahend)
    endif
  endif
  return(difference)
end subtract_bigint_minus_bigint

string proc multiply_bigints(string p_factor1, string p_factor2)
  string  factor1 [MAXSTRINGLEN] = ''
  integer factor1_is_negative    = FALSE
  string  factor2 [MAXSTRINGLEN] = ''
  integer factor2_is_negative    = FALSE
  string  product [MAXSTRINGLEN] = ''
  standardize_bigint(p_factor1, factor1, factor1_is_negative)
  standardize_bigint(p_factor2, factor2, factor2_is_negative)
  if  factor1 <> ''
  and factor2 <> ''
    product = multiply_positive_bigints(factor1, factor2)
    if  product             <> ''
    and factor1_is_negative <> factor2_is_negative
      product = iif(Length(product) == MAXSTRINGLEN, '', '-' + product)
    endif
  endif
  return(product)
end multiply_bigints

string proc divide_bigint_by_bigint(string p_dividend, string p_divisor)
  string  dividend [MAXSTRINGLEN] = ''
  integer dividend_is_negative    = FALSE
  string  divisor  [MAXSTRINGLEN] = ''
  integer divisor_is_negative     = FALSE
  string  quotient [MAXSTRINGLEN] = ''
  standardize_bigint(p_dividend, dividend, dividend_is_negative)
  standardize_bigint(p_divisor , divisor , divisor_is_negative )
  if  dividend <> ''
  and divisor  <> ''
  and divisor  <> '0'
    quotient = divide_positive_bigint_by_bigint(dividend, divisor)
    if  quotient             <> ''
    and dividend_is_negative <> divisor_is_negative
      quotient = iif(Length(quotient) == MAXSTRINGLEN, '', '-' + quotient)
      quotient = iif(quotient == '-0', '0', quotient)
    endif
  endif
  return(quotient)
end divide_bigint_by_bigint

integer proc compare_bigints(string p_left_quantity, string p_right_quantity)
  integer comparison                    = MININT
  string  left_quantity  [MAXSTRINGLEN] = ''
  integer left_quantity_is_negative     = FALSE
  string  right_quantity [MAXSTRINGLEN] = ''
  integer right_quantity_is_negative    = FALSE
  standardize_bigint(p_left_quantity , left_quantity , left_quantity_is_negative )
  standardize_bigint(p_right_quantity, right_quantity, right_quantity_is_negative)
  if  left_quantity  <> ''
  and right_quantity <> ''
    if left_quantity_is_negative == right_quantity_is_negative
      comparison = compare_positive_bigints(left_quantity, right_quantity)
      comparison = iif(left_quantity_is_negative, comparison * -1, comparison)
    else
      comparison = iif(left_quantity_is_negative, -1, 1)
    endif
  endif
  return(comparison)
end compare_bigints

// END OF SAFE FUNCTIONS - THESE CAN BE CALLED BY YOUR OWN CODE

// END OF BIGINT IMPLEMENTATION





// Global constants and semi-constants
#define ALL_FILE_TYPES              -1
string  DIGIT_GROUPING_CHAR   [1] = ','
#define INVALID_HANDLE_VALUE        -1
string  MACRO_NAME [MAXSTRINGLEN] = ''
string  MAXINT_WARNING      [124] = 'This tool currently cannot handle more than 2,147,483,648 files. Report it if the tool aborts on this, and it will be fixed.'


// Global variables
integer inaccessible_directories = 0
integer progress                 = 0
string  windmill             [4] = '/-\|'


string proc group_digits(string size)
  integer p                     = 0
  integer i                     = 0
  string  result [MAXSTRINGLEN] = ''

  for p = Length(size) downto 1
    i = i + 1
    if  i > 1
    and ((i - 1) mod 3) == 0
      result = DIGIT_GROUPING_CHAR + result
    endif
    result = size[p] + result
  endfor

  return(result)
end group_digits


integer proc process_dir(    string  dir_name_in,
                             string  dir_date_time,
                         var string  our_dir_size,
                         var integer our_num_files)
  string  dir_name     [MAXSTRINGLEN] = AddTrailingSlash(dir_name_in)
  integer handle                      = INVALID_HANDLE_VALUE
  integer ok                          = TRUE
  string  sub_dir_size [MAXSTRINGLEN] = '0'
  integer sub_num_files               = 0

  #if DEBUG
    debug_write(dir_name_in + ', ' + dir_date_time)
  #endif

  AddLine(dir_date_time)
  GotoColumn(21)
  InsertText(RemoveTrailingSlash(dir_name))

  our_dir_size  = '0'
  our_num_files =  0

  progress = (progress + 1) mod 4
  Message(windmill[progress + 1], ' ', dir_name_in, ' ... ')

  if not (Lower(dir_name[1:1]) in 'c', 'd', '/')
    // Quick & dirty fix for being too fast for slow network drives.
    Delay(1)
  endif

  if  KeyPressed()
  and GetKey() == <Escape>
    ok = FALSE
  endif

  if ok
    handle = FindFirstFile(dir_name + '*', ALL_FILE_TYPES)

    if handle <> INVALID_HANDLE_VALUE
      repeat
        if (FFAttribute() & _DIRECTORY_)
          if not (FFName() in  '.', '..')
            // "> MAXSTRINGLEN - 2" means "inaccessible to DirSize".
            if  Length(dir_name + FFName()) <= MAXSTRINGLEN - 2
            and FileExists(dir_name + FFName())
              ok = process_dir(dir_name + FFName(),
                               FFDateStr() + ' ' + FFTimeStr(),
                               sub_dir_size,
                               sub_num_files)
              if ok
                our_dir_size  = add_bigints(our_dir_size, sub_dir_size)
                our_num_files = our_num_files + sub_num_files
                if our_num_files < 0
                  Warn(MAXINT_WARNING)
                  ok = FALSE
                endif
              endif
            else
              inaccessible_directories = inaccessible_directories + 1
              AddLine('0 ---------- --------')
              EndLine()
              InsertText(' ')
              InsertText(dir_name)
              InsertText(FFName())
              InsertText(' <inaccessible>')
              Up()
            endif
          endif
        else
          our_dir_size  = add_bigints(our_dir_size, FFSizeStr())
          our_num_files = our_num_files + 1
          if our_num_files < 0
            Warn(MAXINT_WARNING)
            ok = FALSE
          endif
        endif
      until not ok
         or not FindNextFile(handle, ALL_FILE_TYPES)
      FindFileClose(handle)
    endif
  endif

  BegLine()
  InsertText(our_dir_size      , _INSERT_)
  InsertText(' '               , _INSERT_)
  InsertText(Str(our_num_files), _INSERT_)
  InsertText(' '               , _INSERT_)
  Up()

  return(ok)
end process_dir


string proc get_date_time()
  string hh_mm_ss         [8] = GetTimeStr()
  string yyyymmdd_hhmmss [15] = ''
  string yyyy_mm_dd      [10] = GetDateStr()
  yyyymmdd_hhmmss = yyyy_mm_dd[1:4] + yyyy_mm_dd[6:2] + yyyy_mm_dd[9:2] +
                    '_' +
                    hh_mm_ss[1:2] + hh_mm_ss[4:2] + hh_mm_ss[7:2]
  yyyymmdd_hhmmss = StrReplace(' ', yyyymmdd_hhmmss, '0')
  return(yyyymmdd_hhmmss)
end get_date_time


proc WhenPurged()
end WhenPurged


proc WhenLoaded()
  MACRO_NAME = SplitPath(CurrMacroFilename(), _NAME_)
end WhenLoaded


proc Main()
  integer counter                   = 0
  string  dir_date_time        [19] = Query(MacroCmdLine)
  string  dir_name   [MAXSTRINGLEN] = Query(MacroCmdLine)
  string  dir_names  [MAXSTRINGLEN] = Query(MacroCmdLine)
  string  dir_size   [MAXSTRINGLEN] = '0'
  string  list_name  [MAXSTRINGLEN] = Query(MacroCmdLine)
  integer max_size_length           = 0
  integer max_files_length          = 0
  integer num_files                 = 0
  integer ok                        = TRUE
  integer old_dateformat            = Set(DateFormat, 6)  // Sortable.
  integer old_timeformat            = Query(TimeFormat)
  integer total_files               = 0
  string  total_size [MAXSTRINGLEN] = '0'

  if (old_timeformat in 2, 4)
    old_timeformat = Set(TimeFormat, old_timeformat - 1)  // Sortable.
  endif

  if GetFileToken(dir_names, 1) == ''
    if Ask('For which directories do you want to list (sub)directory sizes:',
           dir_names,
           GetFreeHistory(MACRO_NAME + ':DirNames'))
    else
      ok = FALSE
    endif
  endif

  if  ok
  and Lower(Trim(dir_names)) == 'all'
    dir_names = ''
    for counter = Asc('a') to Asc('z')
      if FileExists(Chr(counter) + ':\')
        if dir_names == ''
          dir_names = Chr(counter) + ':'
        else
          dir_names = dir_names + ' ' + Chr(counter) + ':'
        endif
      endif
    endfor
  endif

  if  ok
  and NumFileTokens(dir_names) > 0
    list_name = 'DirSize_' + get_date_time() + '.txt'

    if CreateBuffer(list_name)
      SetUndoOff()
      #if DEBUG
        debug_open()
      #endif
      for counter = 1 to NumFileTokens(dir_names)
        dir_name = AddTrailingSlash(GetFileToken(dir_names, counter))
        if FindThisFile(dir_name + '.')
          dir_date_time = FFDateStr() + ' ' + FFTimeStr()
        elseif Length(dir_name) == 3
        and    isAlpha(dir_name[1])
        and    dir_name[2:2] == ':\'
          dir_date_time = '---------- --------'
        else
          Warn(MACRO_NAME, ': Input directory does not exist:';
               QuotePath(GetFileToken(dir_names, counter)))
          ok = FALSE
        endif
        ok = ok and process_dir(dir_name, dir_date_time, dir_size, num_files)
        if ok
          total_size  = add_bigints(total_size, dir_size)
          total_files = total_files + num_files
          if total_files < 0
            Warn(MAXINT_WARNING)
            ok = FALSE
          endif
        else
          break
        endif
      endfor
      #if DEBUG
        debug_close()
      #endif
      SetUndoOn()
    else
      Warn(MACRO_NAME; 'could not create buffer'; QuotePath(list_name), '.')
      ok = FALSE
    endif
  endif

  if  ok
  and NumLines()
    max_size_length  = Length(group_digits(total_size))
    max_files_length = Length(group_digits(Str(total_files)))
    PushBlock()
    BegFile()

    repeat
      BegLine()
      MarkWord()
      dir_size = GetMarkedText()
      KillBlock()
      InsertText(Format(group_digits(dir_size):max_size_length), _INSERT_)
      WordRight()
      MarkWord()
      num_files = Val(GetMarkedText())
      KillBlock()
      InsertText(Format(group_digits(Str(num_files)):max_files_length), _INSERT_)
    until not Down()

    MarkColumn(1, 1, NumLines(), max_size_length + 1 + max_files_length)
    ExecMacro('sort -d')
    BegFile()
    PopBlock()
  endif

  if Query(Beep)
    Alarm()
  endif
  UpdateDisplay(_ALL_WINDOWS_REFRESH_)

  if ok
    Message(MACRO_NAME; 'ready.')
  else
    Warn(MACRO_NAME; 'aborted.')
  endif

  Set(DateFormat, old_dateformat)
  Set(TimeFormat, old_timeformat)
  PurgeMacro(SplitPath(CurrMacroFilename(), _NAME_))
end

